#!/usr/bin/env python
# encoding: utf-8

from . import convertor

__author__ = 'Liu Yicong'
__email__ = 'imyikong@gmail.com'
__version__ = convertor.__version__
__all__ = ["__author__", "__email__", "__version__", "convertor"]
